
 // Computes empirical k-th order entropy for all k, using a suffix array.
 // It works for all entropies at once (redirect the output!).
 // Context is given by characters following the current one, not preceding.
 // So results can be slightly different from those of entrop and entrop2
 // (look for comment "REVERSE" to achieve the same result of entrop[2])

// #define debug

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

unsigned int *A,*LCP;
unsigned char *T;
unsigned int n;

typedef struct stnode
  { int sdep;
    struct stnode *first,*prev,*next;
  } stree;

int lcp (unsigned char *c1, unsigned char *c2)

  { int d,p = 0;
    while (!(d = *c1++ - *c2++)) p++;
    return p;
  }

void writepar (stree *node, FILE *file)

  { stree *first = node->first;
    fputc('(',file);
    if (first != NULL)
       { node = first;
         do { writepar(node,file);
	      node = node->next;
            }
         while (node != first);
       }
    fputc(')',file);
  }

void main (int argc, char **argv)

  { int t,i,k,m;
    char name[1024];
    FILE *file;
    struct stat sdata;
    float delta;
    unsigned int *PLCP,*IA;
    stree *node;
    stree **S,*M;

    if (argc < 2)
	{ fprintf(stderr,"Usage: %s <file>\n"
			 "       gives all T(k)/k of <file> (to compute delta)\n"
			 "Note: assumes suffix array is in <file>.sa\n",argv[0]);
	  exit(1);
	}

printf ("Reading text\n");

    strcpy(name,argv[1]);
    if (stat (name,&sdata) != 0)
        { fprintf (stderr,"Error: cannot stat file %s\n",name);
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }
     n = sdata.st_size;

    file = fopen(name,"r");
    if (file == NULL)
	{ fprintf(stderr,"Error: cannot open %s for reading.\n",name);
	  fprintf(stderr,"errno = %i\n",errno);
	  exit(1);
	}

     T = (void*)malloc ((n+1) * sizeof(unsigned char));
     if (T == NULL)
        { fprintf (stderr,"Error: not enough memory to build suffix array\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }

    if (fread (T,sizeof(unsigned char),n,file) != n)
	{ fprintf(stderr,"Error: cannot read %s.\n",name);
	  fprintf(stderr,"errno = %i\n",errno);
	  exit(1);
	}
    fclose (file);
    T[n]=0;

#ifdef debug
    printf("T = %s\n",T);
#endif

// Uncomment this to REVERSE the text before computing entropy
// { int j;
//   for (j=0;j<n/2;j++) { int c = T[j]; T[j] = T[n-j-1]; T[n-j-1] = c; }
// }

printf ("Reading suffix array\n");

    strcat(name,".sa");
    file = fopen(name,"r");
    if (file == NULL)
	{ fprintf(stderr,"Error: cannot open %s for reading.\n",name);
	  fprintf(stderr,"errno = %i\n",errno);
	  exit(1);
	}

     A = (void*)malloc ((n+1) * sizeof(unsigned int));
     if (A == NULL)
        { fprintf (stderr,"Error: not enough memory to allocate suffix array\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }

    A[0]=n;
    if (fread (A+1,sizeof(unsigned int),n,file) != n)
	{ fprintf(stderr,"Error: cannot read %s.\n",name);
	  fprintf(stderr,"errno = %i\n",errno);
	  exit(1);
	}
    fclose (file);
    n++;

#ifdef debug
    printf("A = ");
    for (i=0;i<n;i++) printf ("%i, ",A[i]);
    printf("\n");
#endif

printf ("Inverting suffix array\n");

    IA = (void*) malloc (n*sizeof(unsigned int));
    if (IA == NULL)
        { fprintf (stderr,"Error: not enough memory allocate IA array\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }
    for (i=0;i<n;i++) IA[A[i]] = i;

#ifdef debug
    printf("IA = ");
    for (i=0;i<n;i++) printf ("%i, ",IA[i]);
    printf("\n");
#endif

printf ("Computing PLCP\n");

    PLCP = (void*) malloc (n*sizeof(unsigned int));
    if (PLCP == NULL)
        { fprintf (stderr,"Error: not enough memory allocate PLCP array\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }
    if (IA[0] ==0) PLCP[0] = 0;
    else PLCP[0] = lcp(T+A[IA[0]],T+A[IA[0]-1]);
    for (i=1;i<n;i++)
	{ if (IA[i] == 0) PLCP[i] = 0;
	  else if (PLCP[i-1] == 0) PLCP[i] = lcp(T+A[IA[i]],T+A[IA[i]-1]);
	  else PLCP[i] = PLCP[i-1]-1 + lcp(T+A[IA[i]]+PLCP[i-1]-1,
			  		   T+A[IA[i]-1]+PLCP[i-1]-1);
	}

#ifdef debug
    printf("PLCP = ");
    for (i=0;i<n;i++) printf ("%i, ",PLCP[i]);
    printf("\n");
#endif

printf ("Computing LCP\n");

    LCP = IA; // reuse
    for (i=0;i<n;i++) LCP[i] = PLCP[A[i]];
    free (PLCP);
    free(A);

#ifdef debug
    printf("LCP = ");
    for (i=0;i<n;i++) printf ("%i, ",LCP[i]);
    printf("\n");
#endif

printf ("Computing suffix tree\n");

	// stack S, memory M
    S = (void*) malloc (n*sizeof(stree));
    if (S == NULL)
        { fprintf (stderr,"Error: not enough memory allocate stack S\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }
    M = (void*) malloc (2*n*sizeof(struct stnode));
    if (M == NULL)
        { fprintf (stderr,"Error: not enough memory allocate nodes M\n");
          fprintf (stderr," errno = %i\n",errno);
          exit(1);
        }
    m = 0;
    node = M + m++;
    node->sdep = 0;
    node->prev = node->next = node;
    S[0] = node;
	// first leaf for LCP[0]
    node = M + m++;
    node->sdep = -1;
    node->first = NULL;
    node->prev = node->next = node;
    S[0]->first = node;
    k = 0; // top of stack
	// now traverse LCP
    for (i=1;i<n;i++) 
	{ int l = LCP[i];
		// pop sdep > l
	  while (l < S[k]->sdep) k--;
		// case in which we must create an internal node
	  if (l > S[k]->sdep)
            { k++;
	      S[k] = S[k-1]->first->prev;
	      node = M + m++;
	      node->sdep = S[k]->sdep;
  	      node->first = S[k]->first;
	      node->prev = node->next = node;
	      S[k]->sdep = l;
	      S[k]->first = node;
	    }
		// now create the leaf hanging from top of stack
          node = M + m++;
          node->sdep = -1;
          node->first = NULL;
	  S[k]->first->prev->next = node;
 	  node->next = S[k]->first;
	  node->prev = S[k]->first->prev;
	  S[k]->first->prev = node;
	}

printf ("Writing parentheses\n");

    strcpy(name,argv[1]);
    strcat(name,".par");
    file = fopen(name,"w");
    if (file == NULL)
	{ fprintf(stderr,"Error: cannot open %s for writing.\n",name);
	  fprintf(stderr,"errno = %i\n",errno);
	  exit(1);
	}

    writepar(S[0],file);
    fclose(file);

printf ("Done\n");

    exit(0);
  }
